#include <cstdio>
#include <cstdlib>
#include <cmath>

using namespace std;

#define DEFAULT_FREQ 440
#define DEFAULT_VOL  100

static int Stereo;
static int Signed;
static int Bytes;
static int Rate;

enum {Sine,Square,Triangle} Form;

static class Generator *Generators = NULL;

class Generator
{
public:
    class Generator *Next;
public:
    double Tick;
    double DeltaTick;
    double Multi;
    double Volume;
    int Freq;
    
    Generator(Generator *anext, int afreq, int avol) :
        Next(anext), Volume(avol / 100.0), Freq(afreq)
    {
        /* May not setup more here, because the *
         * sampling rate and waveform are not   *
         * necessarily yet configured.          */
    }
    
    void SetupAll(void);    
    void DumpAll(const int _dummy=1);    
    float MixAll(int *_dummy = NULL);
};

void Generator::SetupAll(void)
{
    Multi     = (Form==Sine) ? M_PI+M_PI : 1.0;
    DeltaTick = Multi * (float)Freq / (float)Rate;
    
    Tick      = 0.0;
    
    if(Next)Next->SetupAll( );
}

void Generator::DumpAll(const int _dummy)
{
    if(Next)Next->DumpAll(0);    
    fprintf(stderr, "%d:%d%c", Freq, (int)(0.5+Volume*100), _dummy?'\n':' ');
}
    
float Generator::MixAll(int *_dummy)
{
    register double value = 0.0;
    double temp;
    
    switch(Form)
    {
        case Sine:
            value = sin(Tick);
            break;
        case Square:
            value = modf(Tick, &temp) < 0.5 ? -1 : 1;
            break;
        case Triangle:
            temp = modf(Tick, &temp) * 2.0;
            value = (temp<1.0 ? temp : 2.0 - temp) - 1.0;
            break;
    }
    
    value *= Volume;
    
    Tick += DeltaTick;
    
    if(!_dummy)
    {
        int Count=1;
        
        if(Next)value += Next->MixAll(&Count);
        
        value /= Count;
    }
    else
    {
        (*_dummy)++;
        if(Next)
            value += Next->MixAll(_dummy);
    }
    return value;
}

static int Heelp=0;
static char ProgName[] = "wg";
static void PrintVersionInfo(void)
{
    printf("%s",
        "Wave generator (wg) version " VERSION "\n"
        "Copyright (C) 1992,1999 Bisqwit (http://iki.fi/bisqwit/)\n");
}
static void PrintDescription(void)
{
    printf("\nThis program works like a function generator.\n");
}
static long ParamRange(long min, long max, char **s)
{
    long tmp = strtol(*s, s, 10);
    s--;
    if(tmp<min || tmp>max)
    {
        fprintf(stderr, "%s: parameter value out of range `%ld'\n", ProgName, tmp);
        exit(EXIT_FAILURE);
    }
    return tmp;
}
static void Case(int ID, char **dummy)
{
    switch(ID)
    {
        case 0:    /* startup */
            Stereo = -1;
            Signed = -1;
            Bytes = -1;
            Rate = -1;
            Form = Sine;
            break;
            
        case 'V':
            PrintVersionInfo();
            exit(EXIT_SUCCESS);
            
        case 'r':
            Rate = ParamRange(1, 99999, dummy);
            break;
        case 'b':
            Bytes = ParamRange(1, 2, dummy);
            break;
        case '?': Heelp  = 1; break;
        case 'm': Stereo = 0; break;
        case 's': Stereo = 1; break;
        case 'u': Signed = 0; break;
        case 'i': Signed = 1; break;
        case 'S': Form = Sine; break;
        case 'Q': Form = Square; break;
        case 'T': Form = Triangle; break;        
        case -1: /* finish args, set default values if necessary */
            if(Rate  < 0)Rate  = 44100;
            if(Bytes < 0)Bytes = 2;
            if(Stereo< 0)Stereo= 0;
            if(Signed< 0)Signed = (Bytes==2);
            
            if(!Generators)Generators =
                new Generator(Generators, DEFAULT_FREQ, DEFAULT_VOL);
            break;
    }
    /* Ensure no warnings about unused args */
    dummy = dummy;
}
static struct
{
    int ID;
    char c, c2;
    const char *txt, *descr;
} Options[] =
{
    /* ID-----c--c2----txt--------descr---------------- */
    {'r',    'r',  0,"rate",      "Specify sampling rate. Default: -r44100."},
    {'b',    'b',  0,"bytes",     "Specify byteness... 1=8-bit, 2=16-bit. Default: -b2."},
    {'m',    'm',  0,"mono",      "Specify mono. This is the default."},
    {'s',    's',  0,"stereo",    "Specify stereo. Default: -m."},
    {'u',    'u',  0,"unsigned",  "Specify unsigned samples. This is the default for -b1."},
    {'i',    'i',  0,"signed",    "Specify signed samples. This is the default for -b2."},
    {'S',    'S',  0,"sine",      "Select sine waveform. This is the default."},
    {'Q',    'Q',  0,"square",    "Select square waveform. Default: -S."},
    {'T',    'T',  0,"triangle",  "Select triangle waveform. Default: -S."},
    
    {'?',   '?',  0,"help",      "Help"},
    {'V',   'V',  0,"version",   "Version information"},
    
    {0,0,0,NULL,NULL}
};
static void ArgError(int a, const char *b)
{
    if(a)fprintf(stderr, "%s: unrecognized option `--%s'", ProgName, b);
    else
    {
        fprintf(stderr, "%s: illegal option -- %c", ProgName, *b);
        for(a=0; Options[a].c; a++)
            if(*b == Options[a].c)
            {
                putchar(b[1]);
                break;
            }
    }
    fprintf(stderr, "\nTry `%s --help' for more information.\n", ProgName);
    exit(EXIT_FAILURE);
}
static void ReadArgs(int argc, char **argv)
{
    int a;
    
    Case(0, NULL);

    for(a=1; a<argc; a++)
    {
        char *s = argv[a];

ReArg:  if(!s)continue;
        while(*s==' ' || *s=='\t')s++;
        if(!*s)continue;

        if(*s == '-' || (*s=='/'&&s[1]=='?'))
        {
            if(*++s == '-')
            {
                int b=0;
                for(s++; Options[b].c; b++)
                    if(!strcmp(s, Options[b].txt))
                    {
                        s = argv[++a];
                        Case(Options[b].ID, &s);
                        goto ReArg;
                    }
                ArgError(1, s);
            }
            while(s && *s)
            {
                int b,c;
                for(b=c=0; Options[b].c; b++)
                    if(s[0]==Options[b].c
                    &&(s[1]==Options[b].c2
                          ||!Options[b].c2))
                    {
                        c=1;
                        s++;
                        if(Options[b].c2)s++;
                        Case(Options[b].ID, &s);
                        s--;
                    }
                if(!c)ArgError(0, s);
                if(s)s++;
            }
        }
        else
        {
            int Freq = ParamRange(1, 99999, &s);
            int Vol  = DEFAULT_VOL;
            if(*s==':')
            {
                s++;
                Vol = ParamRange(0, 999, &s);
            }
            Generators = new Generator(Generators, Freq, Vol);
        }
    }
    
    if(Heelp)
    {
        PrintVersionInfo();
        PrintDescription();
        
        printf("\nUsage: %s [-", ProgName);
        for(a=0; Options[a].c; a++)if(!Options[a].c2)printf("%c", Options[a].c);
        printf("] frequency [...]\n\n");

        for(a=0; Options[a].c; a++)
            printf("  -%c%c%c--%-14s%s\n",
                Options[a].c,
                Options[a].c2?Options[a].c2:',',
                Options[a].c2?',':' ',
                Options[a].txt,
                Options[a].descr);

        printf("\n");
        
        exit(EXIT_SUCCESS);
    }

    Case(-1, NULL);
}

int main(int argc, char **argv)
{
    long Limit;
    
    ReadArgs(argc, argv);
    
    fprintf(stderr,
        "%s setup:   %d Hz %d-bit %s (%s) (%s wave)\n"
        "components: ",
        ProgName,
        Rate, Bytes*8, Stereo?"stereo":"mono",
        Signed?"signed":"unsigned",
        Form==Sine?"sine":
            Form==Square?"square":
                Form==Triangle?"triangle":
                    "random"
    );
    
    Generators->DumpAll();
    Generators->SetupAll();
    
    fflush(stderr);
        
    for(Limit = 1 << (8*Bytes-1); ;)
    {
        int a, b;
        long Value, c;

        Value = (long)(Generators->MixAll() * (Limit-1));

        if(Value >= Limit)Value = Limit-1;
        if(Value < -Limit)Value = -Limit;
        
        if(!Signed)Value += Limit;

        for(a=Stereo; a>=0; a--)
            for(c=Value, b=Bytes; --b>=0; )
            {
                putchar(c & 255);
                c >>= 8;
            }
    }
    
    /* never reached, damn. */
    return 0;
}
